# -*- coding: utf-8 -*-
"""
Created on Thu Nov 23 10:21:54 2023

@author: Christophe Bossens
"""

from image_aesthetic_maps.decomposition.quad_tree import QuadTree
from image_aesthetic_maps.composition.mozaique import Mozaique
from image_aesthetic_maps.Patch import Patch
from PIL import Image

import pytest

test_image = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\3-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"

def test_mozaique_returns_image():
    """
    Generating a mozaique should return a PIL image object
    """
    # Arrange
    image = Image.open(test_image)
    root_patch = Patch.from_image(image)
    quadrants = root_patch.decompose(QuadTree)
    
    # Act
    result = Mozaique.generate(image, quadrants)
    
    # Assert
    assert isinstance(result, Image.Image), "Mozaique should generate an image"


def test_mozaique_validates_patches():
    """
    The mozaique function raises an attribute error when patches do not
    have a grid attribute
    """
    # Arrange
    no_attribute_patches = []
    attribute_patches    = []
    for row_index in range(1, 5):
        for col_index in range(1, 5):
            no_attribute_patches.append(Patch(10, 10, 20, 20))
            
            attribute_patch = Patch(10, 10, 30, 30)
            attribute_patch.set_grid_data(row_index, col_index)
            attribute_patches.append(attribute_patch)
            
    # Assert
    with pytest.raises(AttributeError):
        Mozaique.validate_patch_attributes(no_attribute_patches)
        
    assert Mozaique.validate_patch_attributes(attribute_patches) == None, "Should not raise an error"

def test_get_grid_dimensions():
    """
    The mozaique function returns the maximum column and row index for 
    a list of patches

    """
    # Arrange
    MAX_COLUMN_VALUE = 10
    MAX_ROW_VALUE    = 7
    
    patches = []
    for row_index in range(1, MAX_ROW_VALUE + 1):
        for col_index in range(1, MAX_COLUMN_VALUE + 1):
            patch = Patch(10, 10, 30, 30)
            patch.set_grid_data(row_index, col_index)
            patches.append(patch)
            
    # Act
    max_row, max_col = Mozaique.get_maximum_grid_values(patches)
    
    # Assert
    assert max_col == MAX_COLUMN_VALUE, "Maximum column value does not match expected value, %d != %d"%(max_col, MAX_COLUMN_VALUE)
    assert max_row == MAX_ROW_VALUE, "Maximum row value does not match expected value, %d != %d"%(max_row, MAX_ROW_VALUE)
    
    
def test_calculate_new_image_size():
    """
    Should return the appropriate image size
    """
    # Arrange
    original_width = 300
    original_height = 400
    
    n_rows = 5
    n_columns = 7
    
    margin = 10
    
    expected_width = original_width + ((n_columns + 1) * margin)
    expected_height = original_height + ((n_rows + 1) * margin)
    
    # Act 
    new_width, new_height = Mozaique.calculate_new_image_size(original_width, original_height, n_columns, n_rows, margin)
    
    # Assert
    assert new_width == expected_width, "New width does not match expected width: %d != %d"%(new_width, expected_width)
    assert new_height == expected_height, "New height does not match expected width: %d != %d"%(new_height, expected_height)


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 23 10:21:54 2023

@author: u0072088
"""

from image_aesthetic_maps.Patch import Patch
from PIL import Image

test_image = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\3-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"

def test_patch_get_center():
    """
    The function should return the center coordinate of a patch
    """
    # Arrange
    even_dimensions   = Patch(0, 0, 100, 100)
    uneven_dimensions = Patch(0, 0, 101, 101) 
    
    # Act
    patch_center_1 = even_dimensions.get_center()
    patch_center_2 = uneven_dimensions.get_center()
    
    # Assert
    assert patch_center_1[0] == 50 and patch_center_1[1] == 50, "Does not return the correct center for even dimensions"
    assert patch_center_2[0] == 50 and patch_center_1[1] == 50, "Does not return the correct center for uneven dimensions"
    

def test_patch_scale():
    """
    The function should return a new patch by scaling the width and height
    according to the scale factor

    """
    # Arrange
    root_patch = Patch(40, 40, 30, 30)
    
    # Act
    new_patch = root_patch.scale(2)
    
    # Assert
    assert isinstance(new_patch, Patch), "scale() should return a Patch object"
    assert new_patch._width == 60, "Scale should change the width of the new patch"
    assert new_patch._height == 60, "Scale should change the height of the new patch"
    assert new_patch.get_center() == root_patch.get_center(), "Old and new patch should have same center"
    
def test_patch_extract_from_image():
    """
    The function should return an image region from an image.
    

    """
    # Arrange
    root_patch = Patch(40, 40, 30, 30)
    boundary_patch = Patch(-20, -20, 50, 50)
    image = Image.open(test_image)
    
    # Act
    image_patch_1 = root_patch.extract_from_image(image)
    image_patch_2 = boundary_patch.extract_from_image(image)
    
    # Assert
    assert isinstance(image_patch_1, Image.Image), "extract_from_image() should return an Image"
    assert image_patch_1.size == (30, 30), "image size should be equal to patch size"
    assert image_patch_2.size != (50, 50), "Image size should not be equal to patch size"
    
def test_patch_set_grid_data():
    """
    The function should add a _grid attribute to the patch. 
    
    The _grid attribute is a dictionary with a 'col' and 'row' key
    """
    # Arrange 
    patch = Patch(0, 0, 10, 10)
    
    # Act
    patch.set_grid_data(3, 3)
    
    # Assert
    assert hasattr(patch, '_grid'), "The patch does not have a grid attribute"
    assert 'col' in patch._grid, "The _grid dictionary does not have a 'col' property"
    assert 'row' in patch._grid, "The _grid dictionary does not have a 'row' property"
    


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image

from grid_samp import ImageRegion
from grid_samp import ImageRegionList

class RecursiveGrid:
    
    def __init__(self, image: Image.Image | ImageRegion, recursion_depth: int):
        """
        Initialize a RecursiveGrid for recursively subdividing an image or image region.

        This grid subdivides an image into progressively smaller regions according
        to the specified recursion depth. The recursion tree is stored internally,
        with each level containing a list of ImageRegion objects generated from the
        previous level.

        Parameters
        ----------
        image : PIL.Image.Image or ImageRegion
            Source image or an existing ImageRegion to be recursively subdivided.
        recursion_depth : int
            Number of recursive subdivisions to perform. Must be >= 1.

        Raises
        ------
        TypeError
            If `image` is not a PIL.Image.Image or an ImageRegion.
            If `recursion_depth` is not an integer.
        ValueError
            If `recursion_depth` is less than 1.

        Notes
        -----
        If a PIL Image is provided, the first-level ImageRegion is created
        automatically covering the entire image. For each recursion level, the
        `generate` method is used to subdivide the regions from the previous level.

        The resulting tree is stored in `self._tree`, with keys `'level_1'`,
        `'level_2'`, ..., `'level_n'` corresponding to each recursion level.

        Examples
        --------
        >>> from PIL import Image
        >>> from grid_samp.image_region import ImageRegion
        >>> from grid_samp.grids.recursive_grid import RecursiveGrid
        >>> img = Image.open("example.png")
        >>> grid = RecursiveGrid(img, recursion_depth=3)
        >>> grid._recursion_depth
        3
        >>> grid._tree['level_1']  # first-level ImageRegions
        [<ImageRegion>, <ImageRegion>, ...]
        """

        if not isinstance(image, (Image.Image, ImageRegion)):
            raise TypeError("image must be of type PIL.Image.Image or ImageRegion")
        
        if not isinstance(recursion_depth, int):
            raise TypeError("recursion_depth must be of type int")
        
        if recursion_depth < 1:
            raise ValueError("recursion_depth must be at least 1")
        
        self._recursion_depth = recursion_depth
        self._tree = {}

        if isinstance(image, ImageRegion):
            self._source_image = image
        else:
            self._source_image = ImageRegion(0, 0, image.width, image.height)
            self._source_image._grid = {'x': 0, 'y': 0, 'row': 0, 'col': 0, 'depth': 0}
            self._source_image._quadrant = 0

        self._tree['level_1'] = RecursiveGrid.generate(self._source_image)
        for recursion_index in range(2, recursion_depth + 1):
            self._tree[f'level_{recursion_index}'] = []

            for current_level_element in self._tree[f'level_{recursion_index - 1}']:
                self._tree[f'level_{recursion_index}'].extend(RecursiveGrid.generate(current_level_element))

    def get_recursion_level_images(self, recursion_level):
        image_region_list = ImageRegionList(self._tree[f'level_{recursion_level}'])
        sorted_list = sorted(image_region_list, key=lambda image_region: image_region._grid['row'])
        return sorted_list

    @staticmethod
    def generate(image_region):
        width, start_x  = RecursiveGrid.split_length(image_region._width)
        height, start_y = RecursiveGrid.split_length(image_region._height)

        quadrant_1 = ImageRegion(image_region._x0, image_region._y0, width, height)
        quadrant_2 = ImageRegion(image_region._x0 + start_x, image_region._y0, width, height)
        quadrant_3 = ImageRegion(image_region._x0, image_region._y0 + start_y, width, height)
        quadrant_4 = ImageRegion(image_region._x0 + start_x, image_region._y0 + start_y, width, height)

        quadrant_1._quadrant = 1
        quadrant_2._quadrant = 2
        quadrant_3._quadrant = 3
        quadrant_4._quadrant = 4

        src_row = -1
        src_col = -1
        src_depth = 1
        if hasattr(image_region, '_grid'):
            src_row = image_region._grid['row']
            src_col = image_region._grid['col']
            src_depth = image_region._grid['depth']

        quadrant_1._grid = {
            'x': quadrant_1._x0,
            'y': quadrant_1._y0,
            'row': 2 * max(0, src_row) + 0,
            'col': 2 * max(0, src_col) + 0,
            'depth': src_depth + 1
        }

        quadrant_2._grid = {
            'x': quadrant_2._x0,
            'y': quadrant_2._y0,
            'row': 2 * max(0, src_row) + 0,
            'col': 2 * max(0, src_col) + 1,
            'depth': src_depth + 1
        }

        quadrant_3._grid = {
            'x': quadrant_3._x0,
            'y': quadrant_3._y0,
            'row': 2 * max(0, src_row) + 1,
            'col': 2 * max(0, src_col) + 0,
            'depth': src_depth + 1
        }

        quadrant_4._grid = {
            'x': quadrant_4._x0,
            'y': quadrant_4._y0,
            'row': 2 * max(0, src_row) + 1,
            'col': 2 * max(0, src_col) + 1,
            'depth': src_depth + 1
        }

        return [quadrant_1, quadrant_2, quadrant_3, quadrant_4]

    @staticmethod
    def split_length(length):
        """
        For a dimension of 'length' pixels, calculates the width and the starting point
        of the second part when the dimension is split in two.
                
        In the case of even dimensions this works out in two equal sized
        non-overlapping parts
                
        In the case of uneven dimensions, the midlle pixel overlaps in each part.

        Parameters
        ----------
        length : INT
            Number of pixels 

        Returns
        -------
        width : TYPE
            DESCRIPTION.
        start : TYPE
            DESCRIPTION.
        """
        if length % 2 == 0:
            width = length // 2
            start_2 = width
        else:
            width = (length + 1) // 2
            start_2 = width - 1
        return width, start_2



# -*- coding: utf-8 -*-
"""
Created on Thu Nov 23 10:21:54 2023

@author: u0072088
"""

from image_aesthetic_maps.decomposition.quad_tree import QuadTree
from image_aesthetic_maps.Patch import Patch

def test_quad_tree_split_even_length():
    """
    Using the split function with an even number shoud produce a width
    that is equal to the start of the second half
    """
    # Arrange
    test_length = 6
    expected_width = 3
    expected_start = 3
    
    # Act
    width, start = QuadTree.generate_interval_data(test_length)
    
    # Assert
    assert width == expected_width and start == expected_start

def test_quad_tree_split_uneven_length():
    """
    Using the split function with an ueven number should produce a width and a
    start that equals the width - 1
    """
    # Arrange
    test_length = 7
    expected_width = 4
    expected_start = 3
    
    # Act
    width, start = QuadTree.generate_interval_data(test_length)
    
    # Assert
    assert width == expected_width and start == expected_start

def test_quad_tree_equal_patch_sizes():
    """
    Decomposing a patch using the quad tree decomposition should produce
    four equal sized patches
    """
    # Arrange
    patch = Patch(20, 30, 40, 51)
    
    # Act
    quads = QuadTree.generate(patch)
    
    # Assert
    widths = []
    heights = []
    for quad in quads:
        widths.append(quad._width)
        heights.append(quad._height)
        
    assert len(set(widths)) == 1, "All widths should be the same"
    assert len(set(heights)) == 1, "All heights should be the same"
    
def test_quad_tree_stores_decomposition_type():
    """
    Patches generated by the QuadTree decomposition contain an attribute
    that stores the decomposition type
    """
    # Arrange
    patch = Patch(20, 30, 40, 51)
    
    # Act
    patches = QuadTree.generate(patch)
    
    # Assert
    for patch in patches:
        assert hasattr(patch, "_decomposition"), "Patch should have _decomposition attribute"
        assert patch._decomposition == "quad_tree", "Decomposition attribute should equal 'quad_tree'"
        
def test_quad_tree_stores_quadrant_information():
    """
    Patches generated by the QuadTree decomposition contain a _quadrant attribute
    with a unique value from [1, 2, 3, 4]
    """
    # Arrange
    patch = Patch(20, 30, 40, 51)
    
    # Act
    patches = QuadTree.generate(patch)
    
    # Assert
    quadrant_attribute_values = []
    for patch in patches:
        assert hasattr(patch, "_quadrant"), "Patch should have _quadrant attribute"
        quadrant_attribute_values.append(patch._quadrant)
        
    assert len(set(quadrant_attribute_values).intersection([1, 2, 3, 4])) == 4, "There should be four different quadrant attribute values"
    
def test_quad_tree_stores_grid_information():
    """
    Patches generated by the QuadTree decomposition contain a _grid attribute that contains
    the location of each patch in a grid structure.
    These locations are also unique for each of the four patches
    """
    # Arrange
    patch = Patch(20, 30, 40, 51)
    
    # Act
    patches = QuadTree.generate(patch)
    
    # Assert
    grid_attribute_values = []
    for patch in patches:
        assert hasattr(patch, "_grid"), "Patch should have a _grid attribute"
        assert 'row' in patch._grid, "Patch _grid attribute should have a row value"
        assert 'col' in patch._grid, "Patch _grid attribute should have a column value"
        assert 'depth' in patch._grid, "Patch _grid attribute should have depth value"
        grid_attribute_values.append((patch._grid['row'], patch._grid['col']))
        
    assert len(set(grid_attribute_values)) == 4, "Quad tree decomposition should produce four unique grid locations"
    
def test_quad_tree_grid_depth_values():
    """
    When generating a QuadTree, the grid level of the generated patches should
    be one higher than the grid level of the root node
    """
    
    # Arrange
    root_node = Patch(0, 0, 128, 128)
    
    # Act
    level_1_patches = root_node.decompose(QuadTree)
    level_2_patches = level_1_patches[0].decompose(QuadTree)
    
    level_1_depth = [patch._grid['depth'] for patch in level_1_patches]
    level_2_depth = [patch._grid['depth'] for patch in level_2_patches]
    
    # Assert
    assert len(set(level_1_depth)) == 1, "Decomposition should produce a single new depth"
    assert len(set(level_2_depth)) == 1, "Decomposition should produce a single new depth"
    
    assert level_1_depth[0] == level_2_depth[0] - 1, "Depth level should increase for QuadTree decompositions"


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image

from grid_samp import ImageRegion
from grid_samp import ImageRegionList

class RegionBasedGrid:
    def __init__(self, anchor_region: ImageRegion, image: Image.Image):
        """
        Initialize an RegionBasedGrid from an anchor image region.

        This grid is generated by propagating the spatial properties of an
        existing image region (the anchor region) across the given image, 
        dividing the image into equally sized regions based on the anchor.

        Parameters
        ----------
        anchor_region : ImageRegion
            Reference image region that defines the spatial layout of the grid.
            Its position and dimensions are used as the basis for generating
            additional image regions.
        image : PIL.Image.Image
            Source image over which the grid is generated.

        Raises
        ------
        TypeError
            If `anchor_region` is not an instance of ImageRegion.
        ValueError
            If the anchor region does not contain sufficient spatial
            information to generate a grid.

        Notes
        -----
        The generated image regions are computed at initialization time and
        stored internally. The grid is immutable after creation.

        Examples
        --------
        >>> from grid_samp.image_region import ImageRegion
        >>> from grid_samp.grids.image_region_based_grid import RegionBasedGrid
        >>> anchor = ImageRegion(x=0, y=0, width=64, height=64)
        >>> grid = RegionBasedGrid(anchor, image)
        """
        # --- Type validation ---
        if not isinstance(anchor_region, ImageRegion):
            raise TypeError(
                "anchor_region must be an instance of ImageRegion, "
                f"got {type(anchor_region).__name__}"
            )
        
        if not isinstance(image, Image.Image):
            raise TypeError(
                "image must be an instance of Image.Image, "
                f"got {type(image).__name__}"
            )
        
        # --- Spatial validation ---
        required_attributes = ("_width", "_height", "_x0", "_y0")
        missing_attributes = [attribute for attribute in required_attributes if not hasattr(anchor_region, attribute)]

        if missing_attributes:
            raise ValueError(
                "anchor_region does not contain sufficient spatial information "
                "to generate a grid. Missing attributes: "
                f"{', '.join(missing_attributes)}"
            )
        
        self._image_regions = RegionBasedGrid.generate(anchor_region, image)
        
    def get_image_region(self, index):
        return self._image_regions[index]
    
    @property
    def image_region_list(self):
        return ImageRegionList(self._image_regions)
    
    @staticmethod
    def generate(source_Image_Region, image):
        width, height = image.size
        
        width_division = RegionBasedGrid.subdivide(width, source_Image_Region._width, source_Image_Region._x0)
        height_division = RegionBasedGrid.subdivide(height, source_Image_Region._height, source_Image_Region._y0)
        
        image_regions = []
        for row_index, y in enumerate(height_division):
            for col_index, x in enumerate(width_division):
                new_region = ImageRegion(x[0], y[0], x[1], y[1])
                new_region.set_grid_data(row_index + 1, col_index + 1)
                new_region._grid['x'] = x[0]
                new_region._grid['y'] = y[0]
                image_regions.append(new_region)
                
        return image_regions
        
    @staticmethod
    def subdivide(full_size, Image_Region_size, starting_point):
        """
        Generates a list of tuples, containing the starting points and widths
        of each subdivision
        """
        # Get the offset
        offset = starting_point % Image_Region_size
        starting_points = [ [x + offset, Image_Region_size] for x in range(0, full_size, Image_Region_size)]
    
        # Adjust the size of the last element
        if starting_points[-1][0] > full_size:
            starting_points.pop()
            
        if starting_points[-1][0] + starting_points[-1][1] > full_size:
            starting_points[-1][1] = full_size - starting_points[-1][0]
            
        
        # Insert a starting element
        if starting_points[0][0] != 0:
            new_point = [0, starting_points[0][0]]
            starting_points.insert(0, new_point)
                
        return starting_points
        


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 10:39:13 2023

@author: u0072088
"""

from image_aesthetic_maps.decomposition.tile_from_patch import TileFromPatch
from image_aesthetic_maps.Patch import Patch
from PIL import Image

test_image = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\3-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"

def test_returns_list_of_patches():
    patch = Patch(40, 40, 50, 50)
    image = Image.open(test_image)
    
    patches = TileFromPatch.generate(patch, image)
    
    assert isinstance(patches, list), "Return value should be a list"
    for patch in patches:
        assert isinstance(patch, Patch), "List element is not a patch"

def test_subdivide_does_not_return_negative_patch_values():
    patch = Patch(230, 250, 150, 160)
    image = Image.open(test_image)
    
    patches = TileFromPatch.generate(patch, image)
    
    assert isinstance(patches, list), "Return value should be a list"
    for patch in patches:
        assert patch._x0 >= 0, "Value should >= 0"
        assert patch._y0 >= 0, "Value should >= 0"
        assert patch._width >= 0, "Value should >= 0"
        assert patch._height >= 0, "Value should >= 0"
    
def test_subdivide_returns_proper_subdivision_dimension_mismatch():
    """
    The test values should return the corresponding list of elements
    """
    result = TileFromPatch.subdivide(200, 50, 110)
    
    expected_result = [[0, 10], [10, 50], [60, 50], [110, 50], [160, 40]]
    
    assert len(result) == len(expected_result), "Result length does not match expected result"
    
    for idx, patch in enumerate(result):
        assert patch == expected_result[idx], "Result element " + str(patch) + " does not match expected result " + str(expected_result[idx])
        
def test_subdivide_returns_proper_subdivision_dimension_match():
    """
    The test values should return the corresponding list of elements
    """
    result = TileFromPatch.subdivide(200, 50, 100)
    
    expected_result = [[0, 50], [50, 50], [100, 50], [150, 50]]
    
    assert len(result) == len(expected_result), "Result length does not match expected result"
    
    for idx, patch in enumerate(result):
        assert patch == expected_result[idx], "Result element " + str(patch) + " does not match expected result " + str(expected_result[idx])
        
def test_patches_have_unique_grid_attributes():
    """
    The generated patches should have the _grid attribute, and each patch
    should have a unique combination of row and column indices
    """
    # Arrange
    patch = Patch(230, 250, 150, 160)
    image = Image.open(test_image)
    
    # Act
    patches = TileFromPatch.generate(patch, image)
    
    
    # Assert
    for patch in patches:
        assert hasattr(patch, '_grid'), "Patches returned from TileFromPatch have no grid attributes"

    grid_coordinates = [(patch._grid['row'], patch._grid['col']) for patch in patches]
    assert len(set(grid_coordinates)) == len(patches), "Number of unique grid coordinates does not match number of patches"


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image
from grid_samp.assemble.roi import ROI

class Contextualize:
    @staticmethod
    def generate(image, image_region, Image_Region_position = "center", context_position = "right", draw_roi = True):
        """
        Generates a contextualizd image representation of the Image_Region using 
        the provided image

        Parameters
        ----------
        image : PIL Image object
            The full image from which the Image_Region will be extracted.
        Image_Region : Image_Region
            Image_Region object.
        Image_Region_position : STR, optional
            Location where the Image_Region will be drawn, which can be at the center
            or at the original Image_Region location. The default is "center".
        context_position : STR, optional
            Where the original image will be drawn relative to the Image_Region.
            The default is "right".
        draw_roi : BOOL, optional
            Draw a bounding box around the Image_Region in the original image.
            The default is True.

        Returns
        -------
        result : TYPE
            DESCRIPTION.

        """
        # Get the original image dimension
        width, height = image.size

        Image_Region_image = image_region.extract_from_image(image)
        
        if draw_roi:
            image = ROI(image, image_region).draw_region_outline(line_width=5).image
            
        # Ensure both images are in RGBA mode for compatibility
        if image.mode != 'RGBA':
            image = image.convert('RGBA')
        if Image_Region_image.mode != 'RGBA':
            Image_Region_image = Image_Region_image.convert('RGBA')

            
        # Create new image
        if context_position in ["left", "right"]:
            final_height = height
            final_width = 2 * width
        elif context_position in ["top", "bottom"]:
            final_width = width
            final_height = 2 * height
        
        result = Image.new('RGBA', (final_width, final_height))
        
        # generate Image_Region location
        paste_x0 = image_region._x0
        paste_y0 = image_region._y0
        if Image_Region_position == "center":
            center_x = width // 2
            center_y = height // 2
            
            paste_x0 = int(center_x - image_region._width / 2)
            paste_y0 = int(center_y - image_region._height/ 2)
        
        if context_position == "left":
            paste_x0 += width
        if context_position == "top":
            paste_y0 += height
            
        # Paste in original image
        if context_position == "left" or context_position == "top":
            result.paste(image, (0, 0))
        elif context_position == "right":
            result.paste(image, (width, 0))
        elif context_position == "bottom":
            result.paste(image, (0, height))
            
            
        # Paste Image_Region
        result.paste(Image_Region_image, (paste_x0, paste_y0))
        
        return result
        
        
        


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from importlib.metadata import version

__version__ = version("grid-samp")

from grid_samp.image_region import ImageRegion
from grid_samp.image_region_list import ImageRegionList

__all__ = [
    "ImageRegion",
    "ImageRegionList",
]

from grid_samp.grids.fixed_grid import FixedGrid
from grid_samp.assemble.roi import ROI 


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


import copy
import random

class ImageRegionList:
    """
    A class representing a collection of image regions, with options to 
    manipulate the image regions
    """
    
    def __init__(self, image_regions):
        """
        Class constructor

        Parameters
        ----------
        image_regions : list
            A list with ImageRegion objects.

        Returns
        -------
        None.

        """
        self._image_regions = image_regions

    def __repr__(self):
        return "ImageRegionList Object"
        
    def __len__(self):
        """
        Return the number of image regions in the image region list

        """
        return len(self._image_regions)
    
    def __iter__(self):
        """
        Iterator that produces individual ImageRegion objects
        """
        return iter(self._image_regions)
    
    def __getitem__(self, key):
        return self._image_regions[key]
    
    @property
    def image_regions(self):
        """
        Returns the underlying list of ImageRegion objects
        """
        return self._image_regions
        
    def get(self, image_region_index):
        """
        Returns an ImageRegion from the ImageRegionList

        Parameters
        ----------
        image_region_index : int
            Index of the ImageRegion in the ImageRegionList.

        Returns
        -------
        ImageRegion
            A copy of the ImageRegion object.

        """
        return copy.deepcopy(self._image_regions[image_region_index])
        
    def replace(self, original_region_index, new_image_region):
        """
        Returns a copy of the ImageRegionList, in which an image region
        has been replaced with a new image region

        Parameters
        ----------
        original_region_index : int
            Location in the ImageRegionList of the element that will be replaced.
        new_image_region : ImageRegion
            ImageRegion that will be replaced at the provided index.

        Returns
        -------
        ImageRegionList
            A copy of the new ImageRegionList.

        """
        image_region_list = copy.deepcopy(self._image_regions)
        
        if hasattr(image_region_list[original_region_index], '_grid'):
            new_image_region._grid = image_region_list[original_region_index]._grid
        image_region_list[original_region_index] = new_image_region
        
        return ImageRegionList(image_region_list)
        
    def swap(self, region_1_index, region_2_index):
        """
        Returns a copy of the ImageRegionList in which two ImageRegions have
        been swapped

        Parameters
        ----------
        region_1_index : int
            Location of the first image region.
        region_2_index : int
            Location of the second image region.

        Returns
        -------
        ImageRegionList
            A copy of the ImageRegionList in which two regions have been 
            swapped.

        """
        # Create copy for new image region list
        image_regions = copy.deepcopy(self._image_regions)
        
        # Copy the image regions
        region_1_grid = copy.deepcopy(image_regions[region_1_index]._grid)
        region_2_grid = copy.deepcopy(image_regions[region_2_index]._grid)
        
        # Swap the grid properties
        image_regions[region_2_index]._grid = region_1_grid
        image_regions[region_1_index]._grid = region_2_grid
        
        # Return the result
        return ImageRegionList(image_regions)
        
    def shuffle(self, pass_region=None, fix_unit=None):
        regions = copy.deepcopy(self._image_regions)
    
        # Normalize input: convert a single region to a list
        if pass_region is None:
            pass_region = []
            
        elif not isinstance(pass_region, list):
            pass_region = [pass_region]
    
        # Identify indices of regions to exclude
        exclude_indices = []
        for region in pass_region:
            if isinstance(region, int):
                exclude_indices.append(region)
            else:
                for i, r in enumerate(self._image_regions):
                    if r._grid == region._grid:
                        exclude_indices.append(i)
                        break
    
        # Shuffle everything
        print("Ecluding these indices: ",exclude_indices)
        if fix_unit is None:
            all_indices = list(range(len(regions)))
            indices_to_shuffle = [i for i in all_indices if i not in exclude_indices]
            shuffled_indices = indices_to_shuffle.copy()
            random.shuffle(shuffled_indices)
    
            for target_idx, source_idx in zip(indices_to_shuffle, shuffled_indices):
                regions[target_idx]._grid = copy.deepcopy(self._image_regions[source_idx]._grid)
    
        # Shuffle rows
        if fix_unit == 'rows':
            original_row_indices = self._get_grid_row_indices()
            shuffled_row_indices = original_row_indices.copy()
            random.shuffle(shuffled_row_indices)
            grid_row_swaps = {
                orig[0]: new for orig, new in zip(original_row_indices, shuffled_row_indices)
            }
            print("Grid row swaps: ", grid_row_swaps)
    
            for i, image_region in enumerate(regions):
                if i in exclude_indices:
                    continue
                new_grid_location = grid_row_swaps[image_region._grid['row']]
                image_region._grid['row'] = new_grid_location[0]
                image_region._grid['y'] = new_grid_location[1]
                
        # Shuffle columns
        if fix_unit == 'columns':
            original_column_indices = self._get_grid_column_indices()
            shuffled_column_indices = original_column_indices.copy()
            random.shuffle(shuffled_column_indices)
            grid_column_swaps = {
                orig[0]: new for orig, new in zip(original_column_indices, shuffled_column_indices)
            }
            print("Grid row swaps: ", grid_column_swaps)
    
            for i, image_region in enumerate(regions):
                if i in exclude_indices:
                    continue
                new_grid_location = grid_column_swaps[image_region._grid['col']]
                image_region._grid['col'] = new_grid_location[0]
                image_region._grid['x'] = new_grid_location[1]
                
        if fix_unit == "edges":
            all_indices = list(range(len(regions)))
            exclude_indices = self._get_grid_edge_indices()
            indices_to_shuffle = [i for i in all_indices if i not in exclude_indices]
            shuffled_indices = indices_to_shuffle.copy()
            random.shuffle(shuffled_indices)
    
            for target_idx, source_idx in zip(indices_to_shuffle, shuffled_indices):
                regions[target_idx]._grid = copy.deepcopy(self._image_regions[source_idx]._grid)
    
        return ImageRegionList(regions)

    
    def shuffle_within_columns(self):
        return self.shuffle_within_unit('col')

    def shuffle_within_size(self):
        size_groups = {}
        
        # 1. Create groups with equally sized image regions
        for idx, image_region in enumerate(self._image_regions):
            size = (image_region._width, image_region._height)
            
            if not size in size_groups:
                size_groups[size] = [idx]
            else:
                size_groups[size].append(idx)
        
        print(size_groups)
        
        # 2. Shuffle within size groups
        regions = copy.deepcopy(self._image_regions)
        
        for size in size_groups:
            original_indices = size_groups[size]
            shuffled_indices = original_indices.copy()
            random.shuffle(shuffled_indices)
            
            for target_idx, source_idx in zip(original_indices, shuffled_indices):
                regions[target_idx]._grid = copy.deepcopy(self._image_regions[source_idx]._grid)
                
        return ImageRegionList(regions)
        
    
    def shuffle_within_unit(self, unit):
        regions = copy.deepcopy(self._image_regions)
        rows = {}
    
        for idx, region in enumerate(regions):
            row_id = region._grid[unit]
            if row_id not in rows:
                rows[row_id] = []
            rows[row_id].append(idx)
                   
        for row_index in rows:
            row_indices = rows[row_index]
            shuffled_row_indices = copy.deepcopy(row_indices)
            random.shuffle(shuffled_row_indices)
            
            
            for target_idx, source_idx in zip(row_indices, shuffled_row_indices):
                print(target_idx, source_idx)
                regions[target_idx]._grid = copy.deepcopy(self._image_regions[source_idx]._grid)
                       
        return ImageRegionList(regions)
    
    def shuffle_within_rows(self):
        return self.shuffle_within_unit('row')

    
    def _get_grid_row_indices(self):
        """
        Creates a list with all occuring grid row indices ('row') and their
        corresponding position ('y') in the image

        Returns
        -------
        List
            List of tuples: (row_index, y).

        """
        row_indices = []
        for image_region in self._image_regions:
            row_indices.append((image_region._grid['row'], image_region._grid['y']))
        
        return list(set(row_indices))
            
    def _get_grid_column_indices(self):
        """
        Creates a list with all occuring grid column indices ('col') and their
        corresponding position ('x') in the image

        Returns
        -------
        List
            List of tuples: (row_index, y).

        """
        col_indices = []
        for image_region in self._image_regions:
            col_indices.append((image_region._grid['col'], image_region._grid['x']))
            
        return list(set(col_indices))

    
    def _get_grid_edge_indices(self):
        """
        Get the indices of image regions in the list that occur on the edge,
        defined as either the first or last column, or the first or last row

        Returns
        -------
        edge_indices : List
            A list with indices of ImageRegions that fall on the egde of 
            the grid.

        """
        rows = [r._grid['row'] for r in self._image_regions]
        cols = [r._grid['col'] for r in self._image_regions]
    
        min_row, max_row = min(rows), max(rows)
        min_col, max_col = min(cols), max(cols)
    
        edge_indices = []
        for idx, region in enumerate(self._image_regions):
            r, c = region._grid['row'], region._grid['col']
            if r in (min_row, max_row) or c in (min_col, max_col):
                edge_indices.append(idx)
                
        return edge_indices


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 29 13:08:12 2023

@author: u0072088
"""
from PIL import Image

from image_aesthetic_maps.Node import Node
from image_aesthetic_maps.Tree import Tree
from image_aesthetic_maps.decomposition import QuadTree

test_image = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\3-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"


def test_tree_initialization():
    """
    Initializing a tree with a default recursion level should produce a tree
    with one root node and no children
    """
    # Arrange
    image = Image.open(test_image)
    
    # Act
    tree = Tree(image)
    
    # Assert
    assert isinstance(tree._root_node, Node), "Tree has no root node"
    assert isinstance(tree._image, Image.Image), "Tree has no image"
    assert len(tree._root_node._children) == 0, "Default initialization produces no children"
    
def test_get_nodes_at_level():
    """
    Should return a list of nodes at the specified level
    """
    # Arrange
    LEVEL_2_NODES = 3
    LEVEL_3_NODES = 4
    
    image = Image.open(test_image)
    tree = Tree(image)
    
    for i in range(LEVEL_2_NODES):
        child_node = Node()
        tree._root_node.add_child(child_node)
        for j in range(LEVEL_3_NODES):
            child_node.add_child(Node())
            
    # Act
    level_1_nodes = tree._get_nodes_at_level(1)
    level_2_nodes = tree._get_nodes_at_level(2)
    level_3_nodes = tree._get_nodes_at_level(3)
    
    # Assert
    assert len(level_1_nodes) == 1, "Getting the nodes at level 1 should return the root node"
    assert len(level_2_nodes) == LEVEL_2_NODES, "Getting the nodes at level 1 should return the root node"
    assert len(level_3_nodes) == LEVEL_2_NODES * LEVEL_3_NODES, "Getting the nodes at level 1 should return the root node"
    
def test_tree_quad_decomposition():
    """
    Running the tree with a quad decomposition, should produce child nodes that are
    powers of 4
    """
    # Arrange
    image = Image.open(test_image)
    tree = Tree(image)
    
    # Act
    tree.decompose(QuadTree, 3)
    level_1_nodes = tree._get_nodes_at_level(1)
    level_2_nodes = tree._get_nodes_at_level(2)
    level_3_nodes = tree._get_nodes_at_level(3)
    
    # Assert
    assert len(level_1_nodes) == 1, "Getting the nodes at level 1 should return the root node"
    assert len(level_2_nodes) == 4, "Getting the nodes at level 1 should return the root node"
    assert len(level_3_nodes) == 16, "Getting the nodes at level 1 should return the root node"
    
    


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from grid_samp import ImageRegion
from grid_samp import ImageRegionList

class SpacedGrid:
    """
    Generates a grid of equally spaced image regions around 
    a central point, with optional padding for each region.

    Attributes
    ----------
    _image : object
        An image-like object with `.width` and `.height` attributes.
    _x_center : int
        The x-coordinate of the central grid point.
    _y_center : int
        The y-coordinate of the central grid point.
    _x_spacing : int
        Horizontal spacing between the centers of adjacent grid cells.
    _y_spacing : int
        Vertical spacing between the centers of adjacent grid cells.
    _x_padding : int
        Horizontal padding added to each region’s width.
    _y_padding : int
        Vertical padding added to each region’s height.
    _grid_width : int
        Width of the image (derived from `image.width`).
    _grid_height : int
        Height of the image (derived from `image.height`).
    _image_regions : list of ImageRegion
        List of generated image regions in the grid.

    Methods
    -------
    image_regions : ImageRegionList
        Returns an ImageRegionList of the generated grid cells, sorted by row.
    generate_spaced_points(total_size, starting_point, spacing) -> list[int]
        Generates a list of coordinates spaced evenly from a starting point 
        until the edges of the image are reached.
    """
    def __init__(self, image, x_center, y_center, x_spacing, y_spacing, x_padding, y_padding):
        """
        Initializes the SpacedGrid and generates the image regions.

        Parameters
        ----------
        image : object
            An object with `.width` and `.height` attributes (e.g., PIL Image).
        x_center : int
            X-coordinate of the central grid point.
        y_center : int
            Y-coordinate of the central grid point.
        x_spacing : int
            Horizontal distance between centers of adjacent grid cells.
        y_spacing : int
            Vertical distance between centers of adjacent grid cells.
        x_padding : int
            Extra pixels added to each side horizontally.
        y_padding : int
            Extra pixels added to each side vertically.
        """
        self._image = image
        self._x_center = x_center
        self._y_center = y_center
        self._x_spacing = x_spacing
        self._y_spacing = y_spacing
        self._x_padding = x_padding
        self._y_padding = y_padding
        
        self._grid_width  = image.width
        self._grid_height = image.height
        
        self._generate_image_regions()
        
    def _generate_image_regions(self):
        """
        Internal method to generate ImageRegion objects 
        based on the spacing, center, and padding parameters.

        Each region's width and height are determined as:
            width  = 0.5 * x_spacing + (2 * x_padding)
            height = 0.5 * y_spacing + (2 * y_padding)

        The method uses `generate_spaced_points` to determine
        the center positions of all rows and columns, then creates
        ImageRegion objects for each grid cell.

        Notes
        -----
        - This method modifies `self._image_regions` in place.
        - The generated ImageRegion objects are annotated with 
          grid row/column indices and their center coordinates.
        """
        self._image_regions = []
        
        x_centers = SpacedGrid.generate_spaced_points(self._grid_width, self._x_center, self._x_spacing)
        y_centers = SpacedGrid.generate_spaced_points(self._grid_height, self._y_center, self._y_spacing)
        
        for col_index, x in enumerate(x_centers):
            for row_index, y in enumerate(y_centers):
                width  =  int(0.5 * self._x_spacing) + (2 * self._x_padding)
                height =  int(0.5 * self._y_spacing) + (2 * self._y_padding)
                
                print(x, y, width, height)
                
                new_region = ImageRegion(x, y, width, height)
                new_region.set_grid_data(row_index + 1, col_index + 1)
                new_region._grid['x'] = x
                new_region._grid['y'] = y
                self._image_regions.append(new_region)
    
    @property
    def image_regions(self):
        """
        Returns the generated image regions sorted by their row number.

        Returns
        -------
        ImageRegionList
            A list-like container of ImageRegion objects, sorted in 
            top-to-bottom order by row index.
        """
        sorted_list = sorted(self._image_regions, key = lambda image_region : image_region._grid['row'])
        return ImageRegionList(sorted_list)
    
    @staticmethod
    def generate_spaced_points(total_size, starting_point, spacing):
        """
        Generates evenly spaced coordinates along one dimension,
        starting from a given central point and extending toward 
        both edges of the dimension.

        Parameters
        ----------
        total_size : int
            The size of the dimension (e.g., image width or height).
        starting_point : int
            The coordinate of the first central point.
        spacing : int
            The distance between consecutive points.

        Returns
        -------
        list[int]
            A list of coordinates for grid centers, starting from the 
            smallest coordinate (closest to 0) to the largest coordinate 
            (closest to total_size - 1).

        Notes
        -----
        - The first element is `starting_point`.
        - The method fills coordinates in both directions from the 
          starting point until the edges are reached.
        """
        spaced_points = [starting_point]
        
        current_point = starting_point
        while current_point > 0:
            current_point -= spacing
            
            if current_point < 0:
                current_point = 0
                
            spaced_points.insert(0, current_point)
        
        current_point = starting_point
        while current_point < (total_size - 1):
            current_point += spacing
            
            if current_point > (total_size - 1):
                current_point = total_size - 1
                
            spaced_points.append(current_point)
            
        return spaced_points


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image
import random

class Mosaic:
    
    @staticmethod
    def generate(image, image_region_list, margin = 2, linecolor = (255, 255, 255, 0), x_jitter_range = 0, y_jitter_range = 0):
        # Check incoming Image_Regions
        Mosaic.validate_image_region_attributes(image_region_list)
        
        # Get the original image size
        width, height = image.size
        
        # Calculate the size of the Mosaic, taking into margins and offsets
        n_rows, n_cols = Mosaic.get_maximum_grid_values(image_region_list.image_regions)
        total_width, total_height = Mosaic.calculate_new_image_size(width, height, n_rows, n_cols, margin)
        
        # Generate a blank image
        result = Image.new('RGBA', (total_width + x_jitter_range, total_height + y_jitter_range), linecolor)
        
        # Paste the Image_Regions into the Mosaic image
        for image_region in image_region_list.image_regions:
            # Calculate paste coordinates
            x = image_region._grid['x'] + image_region._grid['col'] * margin
            y = image_region._grid['y'] + image_region._grid['row'] * margin
            
            # Adjust for random offset
            x_jitter = random.randint(-x_jitter_range, x_jitter_range)
            y_jitter = random.randint(-y_jitter_range, y_jitter_range)
            
            # Paste Image_Region into final image
            if image_region._image == None:
                extracted_image = image_region.extract_from_image(image)
            else:               
                extracted_image = image_region.extract_from_image(image_region._image)
                
            new_x, new_y = x + x_jitter, y + y_jitter
                
            result.paste(extracted_image, (new_x, new_y))
        
        return result
    
    @staticmethod
    def validate_image_region_attributes(image_region_list):
        """
        To generate a Mosaic, the algorithm needs row and column values
        for the Image_Region. This function checks if these attributes are present
        for each Image_Region.

        Parameters
        ----------
        Image_Regions : list
            a list of Image_Region objects.

        Returns
        -------
        None.

        """
        for image_region in image_region_list.image_regions:
            if not hasattr(image_region, '_grid'):
                raise AttributeError("At least one of the Image_Regions does not have a _grid attribute")
        
    @staticmethod
    def get_maximum_grid_values(Image_Regions):
        max_row = max(Image_Region._grid['row'] for Image_Region in Image_Regions)
        max_col = max(Image_Region._grid['col'] for Image_Region in Image_Regions)
        
        return max_row, max_col

    @staticmethod
    def calculate_new_image_size(original_width, original_height, n_columns, n_rows, margin):
        new_width = original_width + (n_columns + 1) * margin
        new_height = original_height + (n_rows + 1) * margin
        
        return new_width, new_height
        


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image
from grid_samp import ImageRegion
from grid_samp import ImageRegionList

class FixedGrid:
    
    N_ROWS = 2
    N_COLS = 2
    STRICT_DIMENSIONS = True
    
    def __init__(self, image: Image.Image, n_rows: int = 2, n_cols: int = 2, strict_dimensions: bool = False):
        """
        Initialize a FixedGrid for partitioning an image into a regular grid.

        This class divides an input image into a fixed number of rows and
        columns, producing a set of equally sized image regions. The grid
        can optionally enforce strict divisibility of the image dimensions.

        Parameters
        ----------
        image : PIL.Image.Image
            Input image to be partitioned into a fixed grid.
        n_rows : int, optional
            Number of rows in the grid. Must be a positive integer.
            The default is 2.
        n_cols : int, optional
            Number of columns in the grid. Must be a positive integer.
            The default is 2.
        strict_dimensions : bool, optional
            If True, the image width and height must be exactly divisible
            by the number of columns and rows, respectively. If False,
            image regions may have slightly varying dimensions to account
            for remainder pixels. The default is False.

        Raises
        ------
        ValueError
            If `n_rows` or `n_cols` is not a positive integer.
        ValueError
            If `strict_dimensions` is True and the image dimensions are not
            divisible by the specified number of rows and columns.

        Notes
        -----
        The computed image regions are stored internally and can be accessed
        through the grid's image region interface. The grid is immutable
        after creation.

        Examples
        --------
        >>> from PIL import Image
        >>> from grid_samp.grids.fixed_grid import FixedGrid
        >>> img = Image.open("image.png")
        >>> grid = FixedGrid(img, n_rows=4, n_cols=4)
        """
        if not isinstance(image, Image.Image):
            raise TypeError("image must be an instance of PIL.Image.Image")
        
        if not all(isinstance(variable, int) for variable in (n_rows, n_cols)):
            raise TypeError("n_rows and n_cols must be integers")

        if n_rows <= 0 or n_cols <= 0:
            raise ValueError("n_rows and n_cols must be positive integers")
        
        if not isinstance(strict_dimensions, bool):
            raise TypeError("strict_dimensions must be a boolean value")

        FixedGrid.validate_dimensions(image, n_rows, n_cols, strict_dimensions)
        
        self._image = image
        self._n_rows = n_rows
        self._n_cols = n_cols
        
        self._image_regions = FixedGrid.calculate_image_region_coordinates(image, n_rows, n_cols)
        
    
    @property
    def image_regions(self):
        sorted_list = sorted(self._image_regions, key = lambda image_region : image_region._grid['row'])
        return ImageRegionList(sorted_list)
    
    def get_image_region(self, region_index):
        assert region_index < len(self._image_regions), "Region index out of bounds"
        
        return self._image_regions[region_index]
    

    @staticmethod
    def generate(image, n_cols = None, n_rows = None):
        """
        Provides an Image_Region object by decomposing the image in Image_Region
        using the idea of subdivision (the image is divided in HORIZONTAL_PARTS
        along the horizontal dimension and VERTICAL_PARTS along the vertical
        dimension)

        Parameters
        ----------
        image : PIL Image
            PIL Image Object
        n_cols : int
            The number of columns in the image grid
        n_rows : int
            The number of rows in the image grid. Defaults to

        Returns
        -------
        Image_Region collection.

        """
        
        FixedGrid.validate_dimensions(image, n_rows, n_cols)
        
        image_regions = FixedGrid.calculate_image_region_coordinates(image, n_rows, n_cols)
        
        return ImageRegionList(image_regions)

        
    @staticmethod
    def validate_dimensions(image, target_rows, target_columns, strict_dimensions = True):
        """
        Checks if the dimensions of the image are consistent with the 
        HORIZONTAL_PARTS and VERTICAL_PARTS of the Fixed_Grid class

        Parameters
        ----------
        image : PIL Image
            The image for which to check the dimensions.

        Returns
        -------
        None.

        """
        if strict_dimensions:
            assert image.width % target_rows == 0, "Image width needs to be a multiple of %d (currently %d)"%(target_rows, image.width)
            assert image.height % target_columns == 0, "Image height needs to be a multiple of %d (currently %d)"%(target_columns, image.height)
            
        
    @staticmethod
    def calculate_image_region_coordinates(image, n_rows, n_cols):
        """
        Calculates the top left locations and the dimensions for each Image_Region
        when it is split in the configured number of HORIZONTAL_PARTS and 
        VERTICAL_PARTS

        Parameters
        ----------
        image : PIL Image
            A PIL Image object.

        Returns
        -------
        A list with dictionaries with Image_Region data (x, y, width and height)

        """
        # Calculate the Image_Region width and height based on the required number of subdivions
        target_width = image.width//n_cols
        target_height = image.height//n_rows
        
        image_regions = []
        for x in range(n_cols):
            for y in range(n_rows):
                # Calculate the top left corner position of the Image_Region
                x_0 = x * target_width
                y_0 = y * target_height
                
                # Calculate the actual Image_Region width and height, constrained by the actual image size
                region_width = image.width - x_0 if x_0 + target_width > image.width else target_width
                region_height = image.height - y_0 if y_0 + target_height > image.height else target_height
                
                # Create new Image_Region
                image_region = ImageRegion(x_0, y_0, region_width, region_height)
                
                image_region._grid = {
                    "row": y,
                    "col": x,
                    "x": x_0,
                    "y": y_0
                }

                image_regions.append(image_region)
                
                
        return image_regions


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image, ImageDraw, ImageFont, ImageFilter, ImageOps, ImageEnhance, ImageChops

from grid_samp.image_region import ImageRegion
from grid_samp.image_region_list import ImageRegionList

import matplotlib.pyplot as plt
import numpy as np
from skimage import measure
from skimage.exposure import match_histograms

class ROI:
    """
    A utility class for generating ROI objects for region-based image manipulation.

    This class is used to initialize a ROI-aware image from a base image and a defined
    set of regions of interest (ROIs). These regions can be provided as a single
    `ImageRegion` or an `ImageRegionList`.


    Attributes
    ----------
    text_color : tuple
        Default RGB color used for text annotations (default: black).
    text_box_width : int
        Default width of the background rectangle behind ROI labels.
    text_box_height : int
        Default height of the background rectangle behind ROI labels.
    text_box_color : tuple
        Default RGBA color of the background rectangle for text (semi-opaque light gray).
    image : PIL.Image
        The source image to be processed.
    image_regions : list of ImageRegion
        One or more regions of interest defined for the image.
    """

    text_color = (0, 0, 0)
    text_box_width = 20
    text_box_height = 20
    text_box_color = (200, 200, 200, 250)

    def __init__(self, image, image_regions):
        self.image = image.convert("RGB")
        if isinstance(image_regions, (list, ImageRegionList)):
            self.image_regions = list(image_regions)
        elif isinstance(image_regions, ImageRegion):
            self.image_regions = [image_regions]
        else:
            raise ValueError("image_regions must be an ImageRegion or a list/ImageRegionList of ImageRegion objects.")
        self.mode = None

    def __repr__(self):
        return "ROI Object"

##################
# DRAW FUNCTIONS #
##################
    def draw_region_outline(self, line_color="red", line_width=2, draw_index=False):
        """
        Draw bounding boxes or outlines around image regions following their mask contours.
        Falls back to rectangles when no mask is available.
    
        Parameters
        ----------
        line_color : str
        line_width : int
        draw_index : bool
    
        Returns
        -------
        ROI
        """
        draw = ImageDraw.Draw(self.image)
    
        for idx, region in enumerate(self.image_regions):
            if region._mask is not None:
                self._draw_mask_outline(draw, region, line_color, line_width)
            else:
                self._draw_rectangle(draw, region, line_color, line_width)
    
            if draw_index:
                self._draw_region_index(draw, region, idx)
    
        return self
    
    def _draw_rectangle(self, draw, region, line_color, line_width):
        x0, y0 = region.top_left()
        x1, y1 = region.bottom_right()
        draw.rectangle((x0, y0, x1, y1), outline=line_color, width=line_width)
        
    def _draw_region_index(self, draw, region, idx):
        center = region.get_center()
        font = ImageFont.load_default()
        draw.rectangle(
            ((center[0] - 20, center[1] - 20), (center[0] + 20, center[1] + 20)),
            fill=(200, 200, 200, 125)
        )
        draw.text((center[0] - 10, center[1] - 20), str(idx), fill=(0, 0, 0), font=font)
        
    def _draw_mask_outline(self, draw, region, line_color, line_width):
        x0, y0 = region.top_left()
        mask_array = np.array(region._mask)
        binary_mask = mask_array > 0
    
        padded = np.pad(binary_mask, pad_width=1, mode="constant", constant_values=0)
        contours = measure.find_contours(padded.astype(float), 0.5)
    
        if not contours:
            self._draw_rectangle(draw, region, line_color, line_width)
            return
    
        abs_contours = self._contours_to_image_coords(contours, x0, y0)
        self._draw_contour_lines(draw, abs_contours, line_color, line_width)
        self._connect_contour_endpoints(draw, abs_contours, line_color, line_width)
        
    def _contours_to_image_coords(self, contours, offset_x, offset_y, pad=1):
        abs_contours = []
        for contour in contours:
            pts = [(offset_x + pt[1] - pad, offset_y + pt[0] - pad) for pt in contour]
            abs_contours.append(pts)
        return abs_contours
    
    def _draw_contour_lines(self, draw, contours, color, width):
        for pts in contours:
            for i in range(len(pts) - 1):
                draw.line([pts[i], pts[i + 1]], fill=color, width=width)
                
    def _connect_contour_endpoints(self, draw, contours, color, width):
        endpoints = []
        for i, pts in enumerate(contours):
            endpoints.append((pts[0], i, 'start'))
            endpoints.append((pts[-1], i, 'end'))
    
        pairs = []
        used = set()
    
        def dist_sq(p1, p2):
            return (p1[0] - p2[0])**2 + (p1[1] - p2[1])**2
    
        for i in range(len(endpoints)):
            p1, c1, pos1 = endpoints[i]
            for j in range(i + 1, len(endpoints)):
                p2, c2, pos2 = endpoints[j]
                if c1 != c2 and pos1 != pos2:
                    pairs.append((dist_sq(p1, p2), i, j))
    
        for _, i_idx, j_idx in sorted(pairs):
            if i_idx not in used and j_idx not in used:
                p1 = endpoints[i_idx][0]
                p2 = endpoints[j_idx][0]
                draw.line([p1, p2], fill=color, width=width)
                used.add(i_idx)
                used.add(j_idx)
                
########################
# MANIPULATE FUNCTIONS #
########################
    def manipulate(self, mode, alpha_blend_sigma=5):
        """
        Set manipulation mode and falloff for blending manipulated image.
    
        Parameters
        ----------
        mode : str
            "inside" or "outside"
        sigma : float
            Gaussian blur sigma for soft falloff (default: 5)
    
        Returns
        -------
        ROI
        """
        assert mode in ["inside", "outside"], "Mode must be 'inside' or 'outside'"
        self.mode = mode
        self.alpha_blend_sigma = alpha_blend_sigma
        return self


    def _apply_to_mask(self, manipulation_fn, falloff_sigma=0):
        assert self.mode in ["inside", "outside"], "Call .manipulate('inside' | 'outside') before applying."
    
        original = self.image.convert("RGBA")
        manipulated = manipulation_fn(self.image).convert("RGBA")
        mask = Image.new("L", self.image.size, 0)
    
        for region in self.image_regions:
            if hasattr(region, "_mask") and region._mask is not None:
                local_mask = region._mask.convert("L")
                box = (*region.top_left(), *region.bottom_right())
                mask.paste(local_mask, box)
            else:
                x0, y0 = region.top_left()
                x1, y1 = region.bottom_right()
                region_mask = Image.new("L", self.image.size, 0)
                draw = ImageDraw.Draw(region_mask)
                draw.rectangle([x0, y0, x1, y1], fill=255)
                mask = ImageChops.lighter(mask, region_mask)
      
        # Blur mask to get smooth transition
        if self.mode == "outside" and not hasattr(self, "_direction"):
            mask = ImageOps.invert(mask)
            
        blurred_mask = mask.filter(ImageFilter.GaussianBlur(radius=falloff_sigma))
            
        # Blend images using blurred mask
        blended = Image.composite(manipulated, original, blurred_mask)
        blended = blended.convert("RGBA")
        
        if self.mode == "outside":
            if hasattr(self, "_direction"):
                if self._direction == "horizontal":
                    blended = ImageOps.mirror(blended)
                else:
                    blended = ImageOps.flip(blended)
    
        return ROI(blended, self.image_regions)
    
    def _apply_flipping_mask(self, manipulation_fn, falloff_sigma=0):
        assert self.mode in ["inside", "outside"], "Call .manipulate('inside' | 'outside') before applying."
    
        original = self.image.convert("RGBA")
        manipulated = manipulation_fn(self.image).convert("RGBA")
        mask = Image.new("L", self.image.size, 0)
    
        for region in self.image_regions:
            if hasattr(region, "_mask") and region._mask is not None:
                local_mask = region._mask.convert("L")
                box = (*region.top_left(), *region.bottom_right())
                mask.paste(local_mask, box)
            else:
                x0, y0 = region.top_left()
                x1, y1 = region.bottom_right()
                region_mask = Image.new("L", self.image.size, 0)
                draw = ImageDraw.Draw(region_mask)
                draw.rectangle([x0, y0, x1, y1], fill=255)
                mask = ImageChops.lighter(mask, region_mask)
            
        blurred_mask = mask.filter(ImageFilter.GaussianBlur(radius=falloff_sigma))
            
        # Blend images using blurred mask
        blended = Image.composite(manipulated, original, blurred_mask)
        blended = blended.convert("RGBA")
        
        if self.mode == "outside":
            if hasattr(self, "_direction"):
                if self._direction == "horizontal":
                    blended = ImageOps.mirror(blended)
                else:
                    blended = ImageOps.flip(blended)
    
        return ROI(blended, self.image_regions)
    
    def _apply_transparency_mask(self, alpha_fn, alpha: float, falloff_sigma=0):
        assert self.mode in ["inside", "outside"], "Call .manipulate('inside' | 'outside') before applying."
        
        if alpha == 1.0:
            return self

        # Convert to RGBA
        original = self.image.convert("RGBA")
        manipulated = alpha_fn(original.copy())
    
        # Build a combined mask over all regions
        mask = Image.new("L", self.image.size, 0)
        for region in self.image_regions:
            if hasattr(region, "_mask") and region._mask is not None:
                local_mask = region._mask.convert("L")
                box = (*region.top_left(), *region.bottom_right())
                mask.paste(local_mask, box)
            else:
                x0, y0 = region.top_left()
                x1, y1 = region.bottom_right()
                draw = ImageDraw.Draw(mask)
                draw.rectangle([x0, y0, x1, y1], fill=int(alpha * 255))
            print("hello")
            plt.imshow(mask)
    
        if self.mode == "outside":
            mask = ImageOps.invert(mask)
    
        if falloff_sigma > 0:
            mask = mask.filter(ImageFilter.GaussianBlur(radius=falloff_sigma))
    
        # Composite with a fully transparent image
        transparent_base = Image.new("RGBA", self.image.size, (0, 0, 0, 0))
        if alpha < 1.0:
            # Transparent region over visible image
            blended = Image.composite(transparent_base, manipulated, mask)
        else:
            # Opaque region over original image
            blended = Image.composite(manipulated, self.image.convert("RGBA"), mask)
    
        return ROI(blended, self.image_regions)

    def blur(self, radius=5):
        """
        Apply Gaussian blur to the image inside or outside the region(s),
        depending on the mode set by `.manipulate()`.
    
        Parameters
        ----------
        radius : int or float
            Radius of the Gaussian blur.
    
        Returns
        -------
        ROI
        """
        def blur_fn(img):
            return img.filter(ImageFilter.GaussianBlur(radius=radius))
    
        return self._apply_to_mask(blur_fn, falloff_sigma=self.alpha_blend_sigma)
        
    def saturation(self, factor=1.0):
        """
        Adjust color saturation inside or outside the region(s), based on `.manipulate()` mode.
    
        Parameters
        ----------
        factor : float
            Saturation adjustment factor.
            - 1.0 = no change
            - <1.0 = desaturate
            - >1.0 = more saturation
    
        Returns
        -------
        ROIImage
        """
        
        def apply_saturation_fn(img):
            return ImageEnhance.Color(img).enhance(factor)
    
        return self._apply_to_mask(apply_saturation_fn, falloff_sigma=self.alpha_blend_sigma)

    def frequency_filter(self, mode="low", cutoff=40):
        """
        Apply frequency filtering (low-pass or high-pass) to the image in the selected region(s).
    
        Parameters
        ----------
        mode : str
            Type of filter to apply: "low" (low-pass) or "high" (high-pass).
        cutoff : int
            Radius in frequency space to preserve or remove.
        Returns
        -------
        ROIImage
            The modified ROIImage object (self) with frequency filtering applied.
        """
        def apply_freq_filter_to_channel(channel_array):
            f = np.fft.fft2(channel_array)
            fshift = np.fft.fftshift(f)
    
            rows, cols = channel_array.shape
            crow, ccol = rows // 2, cols // 2
    
            y, x = np.ogrid[:rows, :cols]
            mask_area = (x - ccol) ** 2 + (y - crow) ** 2 <= cutoff ** 2
    
            mask = np.zeros_like(channel_array, dtype=np.uint8)
            if mode == "low":
                mask[mask_area] = 1
            elif mode == "high":
                mask[~mask_area] = 1
            else:
                raise ValueError("mode must be 'low' or 'high'")
    
            f_filtered = fshift * mask
            f_ishift = np.fft.ifftshift(f_filtered)
            img_back = np.fft.ifft2(f_ishift)
            return np.abs(img_back).clip(0, 255).astype(np.uint8)
    
        def filter_freq(img):
            if img.mode == "L":
                # Grayscale case
                arr = np.array(img)
                filtered = apply_freq_filter_to_channel(arr)
                return Image.fromarray(filtered).convert("RGB")
            elif img.mode == "RGB":
                arr = np.array(img)
                channels = [apply_freq_filter_to_channel(arr[:, :, i]) for i in range(3)]
                merged = np.stack(channels, axis=2)
                return Image.fromarray(merged.astype(np.uint8))
            else:
                raise ValueError("Unsupported image mode for frequency filtering.")
    
        return self._apply_to_mask(filter_freq, falloff_sigma=self.alpha_blend_sigma)
    
    def flip(self, direction):
        """
        Flip pixels inside or outside the region(s), depending on mode.
        """
        assert direction in ["horizontal", "vertical"], "Direction must be 'horizontal' or 'vertical'"
        assert self.mode in ["inside", "outside"], "Call .manipulate('inside' | 'outside') before flipping."
    
        self._direction = direction
        
        def flip_fn(img):
            return self._region_flip(img, direction, falloff_sigma=self.alpha_blend_sigma)
        
        result = self._apply_flipping_mask(flip_fn, falloff_sigma=self.alpha_blend_sigma)
        
        return result
        
    def _image_flip(self, direction):
        """
        Flip the entire image in the given direction.
        """
        if direction == "horizontal":
            return self.image.transpose(Image.FLIP_LEFT_RIGHT)
        elif direction == "vertical":
            return self.image.transpose(Image.FLIP_TOP_BOTTOM)
    
    def _region_flip(self, img, direction, falloff_sigma=0):
        """
        Flip content inside the ROI, optionally expanding the crop region by falloff_sigma.
        """
        flipped = img.copy()
        w, h = img.size
    
        for region in self.image_regions:
            x0, y0 = region.top_left()
            x1, y1 = region.bottom_right()
    
            # Expand crop box with falloff_sigma, ensuring bounds are within image
            pad = falloff_sigma
            ex0 = max(0, x0 - pad)
            ey0 = max(0, y0 - pad)
            ex1 = min(w, x1 + pad)
            ey1 = min(h, y1 + pad)
    
            region_crop = img.crop((ex0, ey0, ex1, ey1))
    
            if direction == "horizontal":
                region_flipped = region_crop.transpose(Image.FLIP_LEFT_RIGHT)
            elif direction == "vertical":
                region_flipped = region_crop.transpose(Image.FLIP_TOP_BOTTOM)
    
            flipped.paste(region_flipped, (ex0, ey0, ex1, ey1))
    
        return flipped
    
    def set_contrast_reference(self, target_region, target_image):
        """
        Set the reference region and image for histogram contrast matching.

        Parameters
        ----------
        target_region : ImageRegion
            Region in the target image to match contrast to.
        target_image : PIL.Image
            The full target image that contains the target_region.

        Returns
        -------
        None
        """
        self._contrast_match_to = (target_region, target_image)
        return self

    def contrast_match(self):
        """
        Apply histogram matching to adjust the contrast of the region(s)
        inside or outside the mask to match a reference region.
    
        Returns
        -------
        ROIImage
        """
        assert hasattr(self, "_contrast_match_to") and self._contrast_match_to is not None, \
            "Call .set_contrast_reference(target_region, target_image) before contrast_match()"
    
        def apply_contrast_fn(source_image):
            target_region, target_image = self._contrast_match_to
            bbox = target_region.get_bounding_box()
            target_crop = target_image.crop(bbox)
        
            source_crop = source_image.crop(bbox)
        
            source_np = np.array(source_crop)
            target_np = np.array(target_crop)
        
            if np.array_equal(source_np, target_np):
                # Skip matching if they're already equal
                return source_image

            matched_np = match_histograms(
                source_np, target_np,
                channel_axis=-1 if source_np.ndim == 3 else None
            )
            matched_crop = Image.fromarray(np.clip(matched_np, 0, 255).astype(np.uint8))
        
            # Paste matched crop back into original
            result = source_image.copy()
            result.paste(matched_crop, bbox)
            return result
    
        return self._apply_to_mask(apply_contrast_fn, falloff_sigma=self.alpha_blend_sigma)

    def transparency(self, alpha: float = 0.5) -> "ROI":
        def apply_transparency_fn(image: Image.Image) -> Image.Image:
            if image.mode != "RGBA":
                image = image.convert("RGBA")
            r, g, b, _ = image.split()
            new_alpha = Image.new("L", image.size, int(alpha * 255))
            return Image.merge("RGBA", (r, g, b, new_alpha))
    
        return self._apply_transparency_mask(apply_transparency_fn, alpha=alpha, falloff_sigma=self.alpha_blend_sigma)

    def pixel_scramble(self):
        """
        Randomly scramble pixels inside or outside the selected region(s),
        with optional falloff via soft blending.
    
        Returns
        -------
        ROIImage
            A new ROIImage with scrambled pixels blended as per mode and sigma.
        """
    
        def scramble_fn(img):
            arr = np.array(img)
            h, w, c = arr.shape
            flat = arr.reshape(-1, c)
            np.random.shuffle(flat)
            scrambled = flat.reshape(h, w, c)
            return Image.fromarray(scrambled.astype('uint8'))
    
        return self._apply_to_mask(scramble_fn, falloff_sigma=self.alpha_blend_sigma)

    def grey(self):
        """
        Convert the image to greyscale (RGB) in the selected mode.

        Returns
        -------
        ROIImage
            The modified ROIImage object (self) with greyscale applied.
        """
        return self._apply_to_mask(lambda img: ImageOps.grayscale(img).convert("RGB"))
    
    def extract(self):
        """
        Extract ROI image as a PIL Image object.
    
        Returns
        -------
        PIL.Image.Image of the ROI object
        """
        return self.image


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from PIL import Image
import numpy as np
from matplotlib import cm

class ImageRegionIntersectionHeatmap:
    sigma = 50
    alpha = 0.5
    
    @staticmethod
    def generate(image, Image_Regions):
        intersections = ImageRegionIntersectionHeatmap.get_intersection_points(Image_Regions)
        
        x, y = np.meshgrid(np.arange(0, image.width), np.arange(0, image.height))  # Generates empty values at coordinates x and y
        heatmap = np.zeros_like(x, dtype=float)  # Values set to zero
    
        for intersection in intersections:
            center_x, center_y = intersection
    
            heatmap += np.exp(-((x - center_x) ** 2 + (y - center_y) ** 2) / (2.0 * ImageRegionIntersectionHeatmap.sigma ** 2))
    
        heatmap = heatmap / np.max(heatmap)  # Normalize the heatmap to values between 0 and 1
        cmap = cm.get_cmap('viridis')
        rgba_heatmap = cmap(heatmap)
        rgba_heatmap = (rgba_heatmap[:, :, :3] * 255).astype(np.uint8)
        rgb_heatmap = Image.fromarray(rgba_heatmap, 'RGB')
        # heatmap_img = Image.fromarray(heatmap)  # Convert the heatmap to a Pillow image
        
        blended_img = Image.blend(image.convert('RGBA'), rgb_heatmap.convert('RGBA'), alpha = ImageRegionIntersectionHeatmap.alpha)  # Blend the original image with the heatmap
            
        return blended_img
            
            
    def get_intersection_points(Image_Regions):
        corner_count = {}
        for Image_Region in Image_Regions:
            for corner in Image_Region.get_corners():
                if corner not in corner_count:
                    corner_count[corner] = 1
                else:
                    corner_count[corner] += 1
                
        intersections = []
        for corner in corner_count:
            if corner_count[corner] == 4:
                intersections.append(corner)
                
        return intersections


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from grid_samp.assemble.mosaic import Mosaic
from grid_samp.assemble.roi import ROI
from grid_samp.assemble.image_region_intersection_heatmap import ImageRegionIntersectionHeatmap
from grid_samp.assemble.contextualize import Contextualize


# -*- coding: utf-8 -*-
"""
Created on Thu Nov 23 10:21:54 2023

@author: u0072088
"""

from image_aesthetic_maps.decomposition.quad_tree import QuadTree
from image_aesthetic_maps.composition.contextualize import Contextualize
from image_aesthetic_maps.Patch import Patch
from PIL import Image

import pytest

test_image = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\3-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"

def test_contextualize_returns_image():
    """
    Generating a mozaique should return a PIL image object
    """
    # Arrange
    image = Image.open(test_image)
    root_patch = Patch.from_image(image)
    quadrants = root_patch.decompose(QuadTree)
    
    # Act
    result = Contextualize.generate(image, quadrants[1])
    
    # Assert
    assert isinstance(result, Image.Image), "Mozaique should generate an image"




from setuptools import setup, find_namespace_packages

setup(
    name='image_aesthetic_maps',
    version='0.1.0-alpha',
    python_requires = '>=3',
    package_dir = {'' : 'image_aesthetic_maps'},
    packages=find_namespace_packages(where = 'image_aesthetic_maps'),  # Automatically discover and include all packages
    
    # Other metadata
    author='Maarten Leemans',
    author_email='maarten.leemans@kuleuven.be',
    description='A short description of your package',
    long_description='A longer description if needed',
    url='https://github.com/yourusername/mypackage',
    classifiers=[
        # Specify trove classifiers (https://pypi.org/classifiers/)
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)


# grid_samp - Grid-based image sampling toolbox
# Copyright (C) 2026 Maarten Leemans, Crhrisophe Bossens, Johan Wagemans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import copy

from PIL import Image, ImageFilter, ImageDraw

from scipy.fft import fft2, ifft2, fftshift, ifftshift
from skimage.exposure import match_histograms


class ImageRegion:
    def __init__(self, x0:int, y0:int, width:int, height:int, image:Image.Image | None = None):
        """
        Create a new ImageRegion representing a rectangular region of an image.

        Parameters
        ----------
        x0 : int
            The top-left x-coordinate of the region within the image.
        y0 : int
            The top-left y-coordinate of the region within the image.
        width : int
            The width of the region in pixels. Must be a positive integer.
        height : int
            The height of the region in pixels. Must be a positive integer.
        image : PIL.Image.Image, optional
            Optional image from which this region is extracted. If None,
            the region can be associated later with an image. The default is None.

        Raises
        ------
        ValueError
            If `width` or `height` is not positive.

        Notes
        -----
        An ImageRegion stores additional attributes for image processing,
        such as frequency filters, blur, inversion, transparency, greyscale,
        saturation, pixel scrambling, mask, and contrast matching. These
        attributes are initialized to default values and can be modified
        later as needed.

        Examples
        --------
        >>> from PIL import Image
        >>> from grid_samp.image_region import ImageRegion
        >>> img = Image.open("example.png")
        >>> region = ImageRegion(x0=0, y0=0, width=64, height=64, image=img)
        >>> region._width
        64
        >>> region._height
        64
        """
        if not all(isinstance(coordinate, int) for coordinate in (x0, y0, width, height)):
            raise TypeError("x0, y0, width, and height must be integers")
        
        if width <= 0 or height <= 0:
            raise ValueError("width and height must be positive integers")
        
        if image is not None and not isinstance(image, Image.Image):
            raise TypeError("image must be a PIL.Image.Image instance or None")
        
        self._x0 = x0
        self._y0 = y0
        self._width = width
        self._height = height
        self._image = image
        
        self._frequency_filter = None
        self._blur     = False
        self._invert_y = False
        self._invert_x = False
        self._transparency = None
        self._greyscale = False
        self._saturation = 1.0
        self._pixel_scramble = False
        self._mask = None
        self._contrast_match_to = None

        
    def __repr__(self):
        return "ImageRegion Object"
    
    @property
    def mode(self):
        if self._image is not None:
            return self._image.mode
        else:
            return None
        
    def copy(self):
        return copy.deepcopy(self)

    def set_box(self, box):
        self._top_left = (box[0], box[1])
        self._bottom_right = (box[2], box[3])
    
    def get_center(self):
        """
        Returns the center of the ImageRegion. For even dimensions, the center
        is counted as one-half of the way from the origin
        
        Returns
        -------
        A tuple with the center coordinates for the ImageRegion
        
        """
        half_width  = ImageRegion._calculate_center_pixel(self._width)
        half_height = ImageRegion._calculate_center_pixel(self._height)
        
        x_center = self._x0 + half_width
        y_center = self._y0 + half_height
        
        return (x_center, y_center)
    
    def set_image(self, image):
        self._image = image
        
    def get_image(self):
        """
        Returns
        -------
        PIL.Image.Image or None
            The cropped image for this region.
        """
        return self._image
        
    def set_region_dimensions_to_proportions(self, scale_factor):
        """
        Returns a new ImageRegion by scaling the dimensions of the current ImageRegion 
        according to the scale factor, rounded to the nearest integer.
        
        The scaled ImageRegion has the same center as the original ImageRegion.

        Parameters
        ----------
        scale_factor : number or list with two elements
            When a single number is provided, this will scale both dimensions.
            When two numbers are provided, the scaling will be applied to
            the width and height respectively

        Returns
        -------
        ImageRegion.

        """
        if isinstance(scale_factor, int) or isinstance(scale_factor, float):
            scale_factor = [scale_factor, scale_factor]
        
        assert (isinstance(scale_factor, list) or isinstance(scale_factor, tuple)) and len(scale_factor) == 2, "scale factor must be a single number or a list with two numbers"
        
        # Calculate the updated dimensions
        new_width  = int(self._width * scale_factor[0])
        new_height = int(self._height * scale_factor[1])
        
        # Calculate updated origin point centered on current ImageRegion
        new_half_width  = ImageRegion._calculate_center_pixel(new_width)
        new_half_height = ImageRegion._calculate_center_pixel(new_height)
        
        x0 = self.get_center()[0] - new_half_width
        y0 = self.get_center()[1] - new_half_height
        
        # Create and return new ImageRegion
        return ImageRegion(x0, y0, new_width, new_height)
    
    def set_region_dimensions_to_pixels(self, pixel_dims):
        """
        Create a new ImageRegion cropped to a specific pixel width and height,
        centered on the current region.
    
        Parameters
        ----------
        pixel_dims : int, float, or list/tuple of two values
            Desired dimensions of the cropped region in pixels.
            - If a single number, both width and height will use this value.
            - If a list or tuple, should be [width, height].
    
        Returns
        -------
        ImageRegion
            New ImageRegion centered on the original, with given pixel dimensions.
        """
        if isinstance(pixel_dims, (int, float)):
            pixel_dims = [pixel_dims, pixel_dims]
    
        assert isinstance(pixel_dims, (list, tuple)) and len(pixel_dims) == 2, \
            "pixel_dims must be a single number or a list/tuple with two numbers"
    
        new_width = int(pixel_dims[0])
        new_height = int(pixel_dims[1])
    
        # Calculate new top-left corner based on center of current region
        center_x, center_y = self.get_center()
        new_half_width = new_width // 2
        new_half_height = new_height // 2
    
        x0 = center_x - new_half_width
        y0 = center_y - new_half_height
    
        return ImageRegion(x0, y0, new_width, new_height)
    
    def extract_from_image(self, image, clip_to_image_boundaries = True):
        """
        Extract the ImageRegion region from the provided image. ImageRegion coordinates
        are constrained to fall within the image boundaries.

        Parameters
        ----------
        image : PIL image object
            Image from which to extract the ImageRegion region.

        Returns
        -------
        PIL Image
            The ImageRegion extracted from the image.

        """
        if self._frequency_filter:
            image = self._filter_frequency_pass(
                self._frequency_filter["range"],
                self._frequency_filter["cutoff"],
                self._frequency_filter.get("ppd", 30)
                )
        
        # Extract bounding box
        bounding_box = self.get_bounding_box()
        left_bb, top_bb, right_bb, bottom_bb = bounding_box
        width_bb = right_bb - left_bb
        height_bb = bottom_bb - top_bb
    
        # Determine cropped area within image bounds
        crop_left = max(left_bb, 0)
        crop_top = max(top_bb, 0)
        crop_right = min(right_bb, image.width)
        crop_bottom = min(bottom_bb, image.height)
    
        if clip_to_image_boundaries:
            if crop_left != left_bb:
                print(f"WARNING: Left coordinate adjusted from {left_bb} to {crop_left} to stay within image bounds.")
            if crop_top != top_bb:
                print(f"WARNING: Top coordinate adjusted from {top_bb} to {crop_top} to stay within image bounds.")
            if crop_right != right_bb:
                print(f"WARNING: Right coordinate adjusted from {right_bb} to {crop_right} to stay within image bounds.")
            if crop_bottom != bottom_bb:
                print(f"WARNING: Bottom coordinate adjusted from {bottom_bb} to {crop_bottom} to stay within image bounds.")
    
            # Crop and return without transparency padding
            result = image.crop((crop_left, crop_top, crop_right, crop_bottom))
    
        else:
            # Crop the portion within the image bounds
            cropped = image.crop((crop_left, crop_top, crop_right, crop_bottom)).convert("RGBA")
    
            # Create a transparent image the size of the full bounding box
            result = Image.new("RGBA", (width_bb, height_bb), (0, 0, 0, 0))
    
            # Paste the cropped region at the correct offset
            paste_x = crop_left - left_bb
            paste_y = crop_top - top_bb
            result.paste(cropped, (paste_x, paste_y), mask = cropped)
        
        # Perform post processing
        if self._blur:
            result = result.filter(ImageFilter.GaussianBlur(radius = self._blur_radius))
        
        if self._invert_x:
            result = result.transpose(Image.FLIP_LEFT_RIGHT)
            
        if self._invert_y:
            result = result.transpose(Image.FLIP_TOP_BOTTOM)
            
        if self._transparency is not None:
            result = self._apply_transparency(result)
            
        if self._greyscale:
            result = result.convert("L")
            
        if self._saturation != 1.0:
            result = self._adjust_saturation_hsv(result, self._saturation)
            
        if self._pixel_scramble:
            result = self._apply_pixel_scramble(result)
            
        if self._mask is not None:
            # Check mask dimensions match cropped region
            mask_width, mask_height = result.size
            if self._mask.size != (mask_width, mask_height):
                print(f"INFO: Regenerating mask to match clipped region of size {result.size}")
                if hasattr(self, '_generate_oval_mask'):
                    self._mask = self._generate_oval_mask(width=mask_width, height=mask_height)
                else:
                    raise ValueError("Mask and region size mismatch, and no method to regenerate mask.")
            result.putalpha(self._mask)

        if self._contrast_match_to is not None:
            result = self._apply_contrast_match(result)
            
        # Return the result
        return result
    
    def top_left(self):
        """
        Returns the top-left coordinate of the ImageRegion.
        
        Returns
        -------
        tuple
            (x0, y0) coordinate of the top-left corner.
        """
        return (self._x0, self._y0)
    
    def top_right(self):
        """
        Returns the top-right coordinate of the ImageRegion.
        
        Returns
        -------
        tuple
            (x1, y0) coordinate of the top-right corner.
        """
        return (self._x0 + self._width, self._y0)
    
    def bottom_left(self):
        """
        Returns the bottom-left coordinate of the ImageRegion.
        
        Returns
        -------
        tuple
            (x0, y1) coordinate of the bottom-left corner.
        """
        return (self._x0, self._y0 + self._height)
    
    def bottom_right(self):
        """
        Returns the bottom-right coordinate of the ImageRegion.
        
        Returns
        -------
        tuple
            (x1, y1) coordinate of the bottom-right corner.
        """
        return (self._x0 + self._width, self._y0 + self._height)
    
    def decompose(self, decomposition):
        """
        Returns the ImageRegion parameters as a tuple.
        
        Returns
        -------
        tuple
            (x0, y0, width, height)
        """
        return decomposition.generate(self)
    
    def get_corners(self):
        """
        Returns all four corner coordinates of the ImageRegion.
        
        Returns
        -------
        dict
            Dictionary containing 'top_left', 'top_right', 'bottom_left', and 'bottom_right' keys.
        """
        return (self.top_left(), self.top_right(), self.bottom_left(), self.bottom_right())
    
    def get_bounding_box(self):
        """
        Compute the bounding box of the ImageRegion.
        
        Returns
        -------
        tuple
            (left, top, right, bottom) coordinates of the ImageRegion bounding box.
        """
        return (self._x0, self._y0, self._x0 + self._width, self._y0 + self._height)
    
    def set_grid_data(self, row, column):
        """
        Sets the ImageRegion attributes based on grid data.
        
        Parameters
        ----------
        x0 : int
            The x-coordinate of the top-left corner.
        y0 : int
            The y-coordinate of the top-left corner.
        width : int
            The width of the ImageRegion.
        height : int
            The height of the ImageRegion.
        """
        self._grid = {
            'row' : row,
            'col' : column}
        
    def set_mask(self, mask_type, **kwargs):
        """
        Adds a mask when the ImageRegion is extracted from the image.
    
        Parameters
        ----------
        mask_type : int / Image / None
            For integers, the following options are available:
                1: Circular mask
                2: Gaussian mask
            
            If an image is provided, it will be set as the mask.
            
        **kwargs : additional parameters
            sigma: can be used to configure the width of the gaussian mask.
    
        Returns
        -------
        self : ImageRegion
            Returns the modified instance for chaining.
        """
        if mask_type == 1:
            self._mask = self._generate_oval_mask()
        elif mask_type == 2:
            sigma = kwargs.get('sigma', (self._width**2 + self._height**2)**0.5)
            self._mask = self._generate_gaussian_mask(sigma)
        elif isinstance(mask_type, Image.Image):
            assert mask_type.width == self._width and mask_type.height == self._height, 'Mask size does not match region size'
            self._mask = mask_type
        elif mask_type is None:
            self._mask = None
    
        return self

    def _generate_oval_mask(self, width=None, height=None):
        width = width or self._width
        height = height or self._height
        mask = Image.new('L', (width, height), 0)
        draw = ImageDraw.Draw(mask)
        draw.ellipse([(0, 0), (width, height)], fill=255)
        
        return mask

    def _generate_gaussian_mask(self, sigma, width=None, height=None):
    
        width = width or self._width
        height = height or self._height
    
        x = np.linspace(-1, 1, width)
        y = np.linspace(-1, 1, height)
        x, y = np.meshgrid(x, y)
        d = np.sqrt(x*x + y*y)
        g = np.exp(-(d**2) / (2 * sigma**2))
        g = (g / g.max() * 255).astype(np.uint8)
        
        return Image.fromarray(g, mode='L')

       
    def blur(self, status=True, radius=5):
        """
        Enable or disable Gaussian blur for the ImageRegion.
        
        Parameters
        ----------
        status : bool, optional
            If True, the ImageRegion will be blurred. Default is True.
        radius : int, optional
            The blur radius. Default is 5.
        """
        self._blur = status
        self._blur_radius = radius
        
    def invert(self, invert_x = True, invert_y = True):
        """
        Invert the ImageRegion along the x and/or y axis.
        
        Parameters
        ----------
        invert_x : bool, optional
            If True, the ImageRegion will be flipped horizontally. Default is True.
        invert_y : bool, optional
            If True, the ImageRegion will be flipped vertically. Default is True.
        """
        self._invert_x = invert_x
        self._invert_y = invert_y
        
    def set_transparency(self, transparency_level=1.0):
        """
        Set the transparency of the extracted region.
        
        Parameters
        ----------
        transparency_level : float
            A value between 0 (fully transparent) and 1 (fully opaque).
        """
        self._transparency = transparency_level
        
    def _apply_transparency(self, image):
        """
        Apply transparency to the image based on the transparency level.
        """
        image = image.convert("RGBA")
        alpha = image.split()[3]  # Get the alpha channel
        
        # Create a new alpha channel with adjusted transparency
        new_alpha = alpha.point(lambda p: p * self._transparency)
        
        # Replace the alpha channel in the image
        image.putalpha(new_alpha)
        
        return image
    
    def grey_scale(self, status=True):
        """
        Set whether the extracted ImageRegion should be converted to grayscale.
    
        Parameters
        ----------
        status : bool, optional
            If True, the ImageRegion will be converted to grayscale upon extraction. 
            Default is True.
        """
        self._greyscale = status
        
    def set_saturation(self, status=1.0):
        """
        Set the saturation factor. 1.0 means no change, 0.0 means grayscale,
        and >1.0 increases saturation.
        
        Parameters
        ----------
        status : float
            Saturation factor to apply to the ImageRegion. Default is 1.0.
        """
        self._saturation = status
        
    def _adjust_saturation_hsv(self, image, factor):
        """
        Adjusts the saturation of an image using the HSV color space.
    
        Parameters
        ----------
        image : PIL.Image
            Input image in RGB mode.
        factor : float
            Saturation adjustment factor (0 = grayscale, 1 = original, >1 = increased saturation).
    
        Returns
        -------
        PIL.Image
            Image with adjusted saturation.
        """
        # Convert image to HSV
        hsv_image = image.convert("HSV")
        
        # Split the image into H, S, V channels
        h, s, v = hsv_image.split()
        
        # Adjust the saturation (S) channel by multiplying with the factor
        s = s.point(lambda p: min(255, max(0, p * factor)))  # Ensure values are within 0-255
        
        # Merge the channels back into the HSV image
        hsv_image = Image.merge("HSV", (h, s, v))
        
        # Convert back to RGB
        rgb_image = hsv_image.convert("RGB")
        
        return rgb_image
    
    def pixel_scramble(self, status=True):
        """
        Enable or disable pixel scrambling.
        
        Parameters
        ----------
        status : bool
            If True, the pixels will be scrambled. Default is True.
        """
        self._pixel_scramble = status
    
    def _apply_pixel_scramble(self, image):
        """
        Randomly shuffles pixels in the image (pixel scrambling).
        
        Parameters
        ----------
        image : PIL.Image
            The input image to be scrambled.
        
        Returns
        -------
        PIL.Image
            Scrambled image with pixels randomly rearranged.
        """
        # Ensure image is in RGBA to preserve alpha channel if needed
        image = image.convert("RGBA")
        array = np.array(image)
        
        # Flatten the image array to a 2D array of shape (num_pixels, 4)
        h, w, c = array.shape
        flat = array.reshape((-1, c))
        
        # Shuffle pixel rows
        np.random.shuffle(flat)
        
        # Reshape back to the original image shape
        scrambled_array = flat.reshape((h, w, c))
        
        # Convert back to a PIL image
        return Image.fromarray(scrambled_array.astype('uint8'), 'RGBA')
    
    def set_frequency_filter(self, frequency_range, cut_off_frequency, pixels_per_degree=30):
        """
        Set frequency filter parameters for this ImageRegion.
        
        Parameters
        ----------
        frequency_range : tuple or list
            Frequency range to apply the filter.
        cut_off_frequency : float
            Cutoff frequency for the filter.
        pixels_per_degree : int, optional
            Pixels per degree parameter, by default 30.
        
        Returns
        -------
        self
            Returns self for method chaining.
        """
        self._frequency_filter = {
            "range": frequency_range,
            "cutoff": cut_off_frequency,
            "ppd": pixels_per_degree
        }

    
    def _filter_frequency_pass(self, frequency_range="low", cut_off_frequency=2.0, pixels_per_degree=30):
        """
        Applies a low-pass or high-pass spatial frequency filter to a PIL image.
        Preserves the input mode (greyscale or RGB).
    
        Parameters
        ----------
        image : PIL.Image
            Input image to filter.
        frequency_range : str
            'low' for low-pass or 'high' for high-pass filtering.
        cut_off_frequency : float
            Cut-off frequency in cycles per degree.
        pixels_per_degree : float
            Number of pixels per visual degree.
    
        Returns
        -------
        PIL.Image
            Filtered image in the same mode as the input.
        """
        assert frequency_range in ("low", "high"), "frequency_range must be 'low' or 'high'"
        
        if self._image.mode == "L":
            # Greyscale image
            img_array = np.array(self._image, dtype=np.float32)
            rows, cols = img_array.shape
            cx, cy = cols // 2, rows // 2
    
            x = np.linspace(-cx, cx, cols)
            y = np.linspace(-cy, cy, rows)
            xv, yv = np.meshgrid(x, y)
            radius = np.sqrt(xv**2 + yv**2)
            freqs = radius / pixels_per_degree
    
            fft_img = fftshift(fft2(img_array))
            mask = freqs <= cut_off_frequency if frequency_range == "low" else freqs >= cut_off_frequency
            fft_img_filtered = fft_img * mask
            img_filtered = np.real(ifft2(ifftshift(fft_img_filtered)))
    
            # Normalize
            img_filtered -= img_filtered.min()
            img_filtered /= img_filtered.max()
            img_filtered = (img_filtered * 255).astype(np.uint8)
            
            return Image.fromarray(img_filtered, mode="L")
    
        elif self._image.mode == "RGB":
            # Color image: process each channel separately
            channels = self._image.split()
            filtered_channels = []
    
            for ch in channels:
                ch_array = np.array(ch, dtype=np.float32)
                rows, cols = ch_array.shape
                cx, cy = cols // 2, rows // 2
    
                x = np.linspace(-cx, cx, cols)
                y = np.linspace(-cy, cy, rows)
                xv, yv = np.meshgrid(x, y)
                radius = np.sqrt(xv**2 + yv**2)
                freqs = radius / pixels_per_degree
    
                fft_ch = fftshift(fft2(ch_array))
                mask = freqs <= cut_off_frequency if frequency_range == "low" else freqs >= cut_off_frequency
                fft_ch_filtered = fft_ch * mask
                ch_filtered = np.real(ifft2(ifftshift(fft_ch_filtered)))
    
                # Normalize
                ch_filtered -= ch_filtered.min()
                ch_filtered /= ch_filtered.max()
                ch_filtered = (ch_filtered * 255).astype(np.uint8)
                filtered_channels.append(Image.fromarray(ch_filtered))
    
            return Image.merge("RGB", filtered_channels)
    
        else:
            raise ValueError(f"Unsupported image mode: {self._image.mode}")
            

    def _apply_contrast_match(self, source_region: Image.Image) -> Image.Image:
        """
        Apply histogram matching to match the contrast of another region.
    
        Parameters
        ----------
        source_region : PIL.Image
            The cropped image region to be modified.
    
        Returns
        -------
        PIL.Image
            The contrast-matched region.
        """
        if self._contrast_match_to is None:
            return source_region
    
        target_region, target_image = self._contrast_match_to
    
        # Get bounding box and crop target region
        bbox = target_region.get_bounding_box()
        target_crop = target_image.crop(bbox)
    
        # Convert both images to numpy arrays
        source_np = np.array(source_region)
        target_np = np.array(target_crop)
    
        # Perform histogram matching
        matched_np = match_histograms(
            source_np, target_np,
            channel_axis=-1 if source_np.ndim == 3 else None
        )
    
        # Convert back to PIL.Image and return
        return Image.fromarray(np.clip(matched_np, 0, 255).astype(np.uint8))

    def set_contrast_match(self, target_region, target_image):
        """
        Sets the target region and image for contrast matching via histogram.
    
        Parameters
        ----------
        target_region : ImageRegion
            The region whose histogram should be matched.
        target_image : PIL.Image
            The image containing the target region.
        """
        self._contrast_match_to = (target_region, target_image)
    
    @staticmethod
    def _calculate_center_pixel(pixels):
        """
        Calculate the location of the center pixel using zero-based positioning,
        and taking the first half way point in case of uneven dimensions

        Parameters
        ----------
        pixels : INT
            Number of pixels

        Returns
        -------
        INT
            Location of the center pixel.

        """
        return pixels // 2 if pixels % 2 == 0 else (pixels - 1) // 2
    
    @staticmethod
    def from_image(image):
        """
        Set up an initial ImageRegion based on the dimensions of the provided image

        Parameters
        ----------
        image : PIL Image
        

        Returns
        -------
        ImageRegion : ImageRegion object
            
        """
        
        image_region = ImageRegion(0, 0, image.width, image.height, image)
        
        return image_region
    
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    
    image_file = r"C:\Users\u0072088\OneDrive - KU Leuven\Documents\03-projects\image-aesthetic-map-toolbox\image-aesethic-map-toolbox\tests\test_image.jpg"
    image = Image.open(image_file)
    
    image_region = ImageRegion.from_image(image)
    scrambled_image = image_region.pixel_scramble(image)
    plt.imshow(scrambled_image)